#
# ademco.system - Security systems
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import deque

import asyn
import asyn.utility

import core
import event

DEBUG = None


#
# Constants
#
IDLE_PROBE = '..'		# illegal command used for idle probes


#
# A queued outgoing command
#
class Command(core.SystemPart, asyn.Callable):
	def __init__(self, system, cmd, callout=None):
		core.SystemPart.__init__(self, system)
		asyn.Callable.__init__(self, callout=callout)
		self.cmd = cmd


#
# A System Access Module
#
class System(asyn.Callable, asyn.utility.Idler):
	""" A security system interface.

		The interface represents (exactly) one security system,
		so we might as well call it that.
	"""
	def __init__(self, control, callout=None):
		asyn.Callable.__init__(self, callout=callout)
		asyn.utility.Idler.__init__(self, control)
		self.control = control
		self._io = None
		self._wqueue = deque()
		self._active = None
		self.partition = { }
		self.zone = { }

	SCAN = asyn.scan.Regex([
		(r'(.*)\r\n', 'line')		# parse into text lines
	])


	def open(self, file):
		""" Establish contact with the security system.

			The file argument must be based on a file descriptor - it can be either
			a genuine File object, or an integer file descriptor. Both sockets and
			local streams are acceptable.
		"""
		self._io = self.control.stream(file, callout=self._event)
		self._io.read_flush(discard=True)
		self._io.scan = self.SCAN
		self._discovered = False	# guard against early notifications

		# fill command queue with initial discovery program
		self.command(IDLE_PROBE)	# wakeup (purge channel)
		self.command('ZP')			# zone/partition discovery
		self.zone_status()			# get zone status
		self.arming_status()		# get partition status


	#
	# Close down the object
	#
	def close(self):
		""" Release system resources. """
		self.idle_cancel()
		if self._io:
			self._io.close()
			self._io = None


	#
	# Make partitions or zones as needed. Used by discovery.
	#
	def make_partition(self, pn):
		part = self.partition.get(pn)
		if part is None:
			part = self.partition[pn] = core.Partition(self, pn)
			self.callout('discover', 'partition', part)
		return part

	def make_zone(self, zn, part):
		zone = self.zone.get(zn)
		if zone is None:
			zone = self.zone[zn] = core.Zone(self, zn, part)
		return zone


	#
	# Misc. requests
	#
	def read_log(self):
		""" Trigger log read-out. Log entries are delivered as events. """
		self.command('LD')


	#
	# Idle trigger
	#
	def idle(self):
		""" Issue a no-op command to check connectivity. """
		self.command(IDLE_PROBE)

	def idle_timeout(self):
		self.callout("timeout")


	#
	# Arming and disarming
	#
	def arm(self, type, user, partlist=None):
		self._sendArm(type, user, partlist)

	def disarm(self, user, partlist=None):
		self._sendArm('AD', user, partlist)

	def _sendArm(self, cmd, user, partlist):
		if partlist is None:
			partlist = '1'
		elif isinstance(partlist, basestring):
			pass
		else:
			try:
				partlist = ''.join([str(part.pn) for part in partlist])
			except TypeError:
				partlist = str(partlist)	# single
		self.command("%s%s%s%s" % (
			cmd,
			"%02.2d" % user.uid,
			user.pin,
			partlist.zfill(8)
		))


	#
	# Keypad control
	#
	def keypad_type(self, partition, chars):
		maxper=5
		for i in xrange(0, len(l), maxper):
			self.command('KS%s%s' % (partition.pn, chars[i:i+maxper]))


	#
	# Update drivers
	#
	def zone_status(self):
		self.command('ZS')		# zone status

	def arming_status(self):
		self.command('as')		# arming status


	#
	# Update operations
	#
	def updateZoneMap(self, body):
		# PZ....Z
		row = int(body[0])
		assert row in [1,2,3,4]
		base = 64 * (row - 1)	# zone number of first indicator
		for z in range(1, len(body)):
			zn = base + z		# zone number
			pn = int(body[z])	# assigned to partition
			if pn:
				self.make_zone(zn, self.make_partition(pn))


	def updateZoneStatus(self, body):
		# MD....D
		row = int(body[0])
		base = 64 * (row - 1)
		for z in range(1, len(body)):
			zn = base + z
			state = int(body[z], 16)
			if zn in self.zone:
				self.zone[zn].updateStatus(state)
			elif state:
				print "Zone %d status %d not in partition map" % (zn, state)

	def updateArmingStatus(self, body):
		assert len(body) == core.MAX_PARTITIONS
		for pp in range(core.MAX_PARTITIONS):
			pn = pp+1
			if pn in self.partition:
				self.partition[pn].updateStatus(body[pp])
		self._discovered = True			# that was the last one
		self.callout('discover', 'done')


	def updateZoneDescriptor(self, body):
		# zzz"description"
		assert body[3] == '"' and body[-1] == '"'
		zn = int(body[0:3])
		self.zone[zn].description = body[4:-1]


	def updateState(self, ev):
		if not self._discovered:
			return			# ignore early update notifications
		if ev.code == event.FAULT:
			self.zone[ev.zone].updateStatus(core.ZS_OPEN)
		elif ev.code == event.UNFAULT:
			self.zone[ev.zone].updateStatus(core.ZS_CLOSED)
		elif ev.code == event.ARM_AWAY:
			self.partition[ev.partition].updateStatus(core.ARM_AWAY)
			self.command('ZS')
		elif ev.code == event.ARM_STAY:
			self.partition[ev.partition].updateStatus(core.ARM_STAY)
			self.command('ZS')
		elif ev.code == event.DISARM:
			self.partition[ev.partition].updateStatus(core.DISARM)
			self.command('ZS')


	#
	# Omnibus event handler.
	# The asyn parser only separates lines; we do the rest here.
	#
	def _event(self, ctx, line=None):
		if ctx.error:
			return self.callout(ctx)
		if ctx.state == 'CLOSE':	# connection lost
			self._io = None			# any retry to be done by upstream
			return self.callout('close') # ... so let them know
		if ctx.state == 'END':
			return self.callout(ctx)
		if ctx.state != 'line':
			print 'UNEXPECTED EVENT TYPE', ctx
			return
		if DEBUG: DEBUG('<-', ctx, line)
		# unframe and check
		msg = core.unframe(line)
		if msg is None:
			return self.callout('invalid', line)
		else:
			self.callout('received', *msg)
		# dispatch
		(code, body) = msg
		if code == 'XN':
			self.callout('communications', 'off')
		elif code == 'XF':
			self.callout('communications', 'on')
		elif code == 'FV':
			if body != IDLE_PROBE:
				self.callout('ack', body)
		elif code == 'OK':
			self._end_xmit()
		elif code == 'nq':
			ev = event.SystemEvent(self, body)
			self.updateState(ev)
			self.callout('event', ev)
		elif code == 'ld':
			self.callout('log event', event.SystemEvent(self, body))
		elif code == 'lc':
			self._end_xmit()
			self.callout('log end')
		elif code == 'AS':
			self.updateArmingStatus(body)
		elif code == 'ZP':
			self.updateZoneMap(body)
		elif code == 'ZS':
			self.updateZoneStatus(body)
		elif code == 'zd':
			if body == '000""':
				self._end_xmit()
				self.callout('end descriptors')
			else:
				self.updateZoneDescriptor(body)
		elif code == 'kd':
			# send this (only) to the requesting caller
			self._active.callout('keypad', body)
		else:
			print 'unexpected event code', code
		self.idle_activity()


	#
	# Command driver
	#
	def command(self, cmd, callout=None):
		self._wqueue.append(Command(self, cmd, callout=callout))
		self._xmit()

	def _xmit(self):
		if self._active is None:
			try:
				self._active = self._wqueue.popleft()
				self._write(self._active.cmd)
			except IndexError:
				return

	def _end_xmit(self):
		if self._active:
			self._active.callout('done', self._active)
		self._active = None
		self._xmit()

	def _write(self, cmd):
		msg = core.frame(cmd)
		msg += core.cksum(msg)
		if DEBUG: DEBUG('->', msg)
		self._io.write(msg + '\r\n')
